@@ -31,16 +31,11 @@
 import struct
 import zlib
 from abc import abstractmethod
+from collections.abc import Generator, Iterable, Iterator, Mapping
 from datetime import datetime
 from typing import (
     Any,
-    Dict,
-    Iterable,
-    Iterator,
-    List,
-    Mapping,
     Optional,
-    Tuple,
     Union,
     cast,
 )
@@ -49,7 +44,7 @@ from ._encryption import Encryption
 from ._page import PageObject, _VirtualList
 from ._page_labels import index2label as page_index2page_label
 from ._utils import (
-    deprecate_with_replacement,
+    deprecation_with_replacement,
     logger_warning,
     parse_iso8824_date,
 )
@@ -58,13 +53,13 @@ from .constants import CatalogDictionary as CD
 from .constants import (
     CheckboxRadioButtonAttributes,
     GoToActionArguments,
+    PagesAttributes,
     UserAccessPermissions,
 )
 from .constants import Core as CO
 from .constants import DocumentInformationAttributes as DI
 from .constants import FieldDictionaryAttributes as FA
 from .constants import PageAttributes as PG
-from .constants import PagesAttributes as PA
 from .errors import PdfReadError, PyPdfError
 from .generic import (
     ArrayObject,
@@ -87,13 +82,14 @@ from .generic import (
     create_string_object,
     is_null_or_none,
 )
+from .generic._files import EmbeddedFile
 from .types import OutlineType, PagemodeType
 from .xmp import XmpInformation
 
 
-def convert_to_int(d: bytes, size: int) -> Union[int, Tuple[Any, ...]]:
+def convert_to_int(d: bytes, size: int) -> Union[int, tuple[Any, ...]]:
     if size > 8:
-        raise PdfReadError("invalid size in convert_to_int")
+        raise PdfReadError("Invalid size in convert_to_int")
     d = b"\x00\x00\x00\x00\x00\x00\x00\x00" + d
     d = d[-8:]
     return struct.unpack(">q", d)[0]
@@ -121,6 +117,8 @@ class DocumentInformation(DictionaryObject):
         retval = self.get(key, None)
         if isinstance(retval, TextStringObject):
             return retval
+        if isinstance(retval, ByteStringObject):
+            return str(retval)
         return None
 
     @property
@@ -241,6 +239,21 @@ class DocumentInformation(DictionaryObject):
         """
         return self.get(DI.MOD_DATE)
 
+    @property
+    def keywords(self) -> Optional[str]:
+        """
+        Read-only property accessing the document's keywords.
+
+        Returns a ``TextStringObject`` or ``None`` if keywords are not
+        specified.
+        """
+        return self._get_text(DI.KEYWORDS)
+
+    @property
+    def keywords_raw(self) -> Optional[str]:
+        """The "raw" version of keywords; can return a ``ByteStringObject``."""
+        return self.get(DI.KEYWORDS)
+
 
 class PdfDocCommon:
     """
@@ -251,6 +264,8 @@ class PdfDocCommon:
 
     strict: bool = False  # default
 
+    flattened_pages: Optional[list[PageObject]] = None
+
     _encryption: Optional[Encryption] = None
 
     _readonly: bool = False
@@ -299,22 +314,6 @@ class PdfDocCommon:
     def xmp_metadata(self) -> Optional[XmpInformation]:
         ...  # pragma: no cover
 
-    @abstractmethod
-    def _repr_mimebundle_(
-        self,
-        include: Union[None, Iterable[str]] = None,
-        exclude: Union[None, Iterable[str]] = None,
-    ) -> Dict[str, Any]:
-        """
-        Integration into Jupyter Notebooks.
-
-        This method returns a dictionary that maps a mime-type to its
-        representation.
-
-        See https://ipython.readthedocs.io/en/stable/config/integrating.html
-        """
-        ...  # pragma: no cover
-
     @property
     def viewer_preferences(self) -> Optional[ViewerPreferences]:
         """Returns the existing ViewerPreferences as an overloaded dictionary."""
@@ -324,14 +323,12 @@ class PdfDocCommon:
         o = o.get_object()
         if not isinstance(o, ViewerPreferences):
             o = ViewerPreferences(o)
-            if hasattr(o, "indirect_reference"):
+            if hasattr(o, "indirect_reference") and o.indirect_reference is not None:
                 self._replace_object(o.indirect_reference, o)
             else:
                 self.root_object[NameObject(CD.VIEWER_PREFERENCES)] = o
         return o
 
-    flattened_pages: Optional[List[PageObject]] = None
-
     def get_num_pages(self) -> int:
         """
         Calculate the number of pages in this PDF file.
@@ -340,19 +337,18 @@ class PdfDocCommon:
             The number of pages of the parsed PDF file.
 
         Raises:
-            PdfReadError: if file is encrypted and restrictions prevent
-                this action.
+            PdfReadError: If restrictions prevent this action.
+
         """
         # Flattened pages will not work on an encrypted PDF;
         # the PDF file's page count is used in this case. Otherwise,
         # the original method (flattened page count) is used.
         if self.is_encrypted:
             return self.root_object["/Pages"]["/Count"]  # type: ignore
-        else:
-            if self.flattened_pages is None:
-                self._flatten(self._readonly)
-            assert self.flattened_pages is not None
-            return len(self.flattened_pages)
+        if self.flattened_pages is None:
+            self._flatten(self._readonly)
+        assert self.flattened_pages is not None
+        return len(self.flattened_pages)
 
     def get_page(self, page_number: int) -> PageObject:
         """
@@ -365,6 +361,7 @@ class PdfDocCommon:
 
         Returns:
             A :class:`PageObject<pypdf._page.PageObject>` instance.
+
         """
         if self.flattened_pages is None:
             self._flatten(self._readonly)
@@ -374,7 +371,7 @@ class PdfDocCommon:
     def _get_page_in_node(
         self,
         page_number: int,
-    ) -> Tuple[DictionaryObject, int]:
+    ) -> tuple[DictionaryObject, int]:
         """
         Retrieve the node and position within the /Kids containing the page.
         If page_number is greater than the number of pages, it returns the top node, -1.
@@ -383,17 +380,15 @@ class PdfDocCommon:
 
         def recursive_call(
             node: DictionaryObject, mi: int
-        ) -> Tuple[Optional[PdfObject], int]:
+        ) -> tuple[Optional[PdfObject], int]:
             ma = cast(int, node.get("/Count", 1))  # default 1 for /Page types
             if node["/Type"] == "/Page":
                 if page_number == mi:
                     return node, -1
-                # else
                 return None, mi + 1
             if (page_number - mi) >= ma:  # not in nodes below
                 if node == top:
                     return top, -1
-                # else
                 return None, mi + ma
             for idx, kid in enumerate(cast(ArrayObject, node["/Kids"])):
                 kid = cast(DictionaryObject, kid.get_object())
@@ -401,7 +396,7 @@ class PdfDocCommon:
                 if n is not None:  # page has just been found ...
                     if i < 0:  # ... just below!
                         return node, idx
-                    # else:  # ... at lower levels
+                    # ... at lower levels
                     return n, i
                 mi = i
             raise PyPdfError("Unexpectedly cannot find the node.")
@@ -411,11 +406,8 @@ class PdfDocCommon:
         return node, idx
 
     @property
-    def named_destinations(self) -> Dict[str, Any]:
-        """
-        A read-only dictionary which maps names to
-        :class:`Destinations<pypdf.generic.Destination>`
-        """
+    def named_destinations(self) -> dict[str, Destination]:
+        """A read-only dictionary which maps names to destinations."""
         return self._get_named_destinations()
 
     def get_named_dest_root(self) -> ArrayObject:
@@ -424,9 +416,8 @@ class PdfDocCommon:
             self.root_object[CA.NAMES], DictionaryObject
         ):
             names = cast(DictionaryObject, self.root_object[CA.NAMES])
-            names_ref = names.indirect_reference
             if CA.DESTS in names and isinstance(names[CA.DESTS], DictionaryObject):
-                # 3.6.3 Name Dictionary (PDF spec 1.7)
+                # §3.6.3 Name Dictionary (PDF spec 1.7)
                 dests = cast(DictionaryObject, names[CA.DESTS])
                 dests_ref = dests.indirect_reference
                 if CA.NAMES in dests:
@@ -456,18 +447,18 @@ class PdfDocCommon:
     def _get_named_destinations(
         self,
         tree: Union[TreeObject, None] = None,
-        retval: Optional[Any] = None,
-    ) -> Dict[str, Any]:
+        retval: Optional[dict[str, Destination]] = None,
+    ) -> dict[str, Destination]:
         """
         Retrieve the named destinations present in the document.
 
         Args:
-            tree:
-            retval:
+            tree: The current tree.
+            retval: The previously retrieved destinations for nested calls.
 
         Returns:
-            A dictionary which maps names to
-            :class:`Destinations<pypdf.generic.Destination>`.
+            A dictionary which maps names to destinations.
+
         """
         if retval is None:
             retval = {}
@@ -481,21 +472,22 @@ class PdfDocCommon:
                 if CA.DESTS in names:
                     tree = cast(TreeObject, names[CA.DESTS])
 
-        if tree is None:
+        if is_null_or_none(tree):
             return retval
+        assert tree is not None, "mypy"
 
-        if PA.KIDS in tree:
+        if PagesAttributes.KIDS in tree:
             # recurse down the tree
-            for kid in cast(ArrayObject, tree[PA.KIDS]):
+            for kid in cast(ArrayObject, tree[PagesAttributes.KIDS]):
                 self._get_named_destinations(kid.get_object(), retval)
         # §7.9.6, entries in a name tree node dictionary
         elif CA.NAMES in tree:  # /Kids and /Names are exclusives (§7.9.6)
             names = cast(DictionaryObject, tree[CA.NAMES])
             i = 0
             while i < len(names):
-                key = cast(str, names[i].get_object())
+                key = names[i].get_object()
                 i += 1
-                if not isinstance(key, str):
+                if not isinstance(key, (bytes, str)):
                     continue
                 try:
                     value = names[i].get_object()
@@ -507,9 +499,11 @@ class PdfDocCommon:
                         value = value["/D"]
                     else:
                         continue
-                dest = self._build_destination(key, value)  # type: ignore
+                dest = self._build_destination(key, value)
                 if dest is not None:
-                    retval[key] = dest
+                    retval[cast(str, dest["/Title"])] = dest
+                    # Remain backwards-compatible.
+                    retval[str(key)] = dest
         else:  # case where Dests is in root catalog (PDF 1.7 specs, §2 about PDF 1.1)
             for k__, v__ in tree.items():
                 val = v__.get_object()
@@ -523,16 +517,16 @@ class PdfDocCommon:
                     retval[k__] = dest
         return retval
 
-    # A select group of relevant field attributes. For the complete list.
-    # See §12.3.2 of the PDF 1.7 or PDF 2.0 specification.
+    # A select group of relevant field attributes. For the complete list,
+    # see §12.3.2 of the PDF 1.7 or PDF 2.0 specification.
 
     def get_fields(
         self,
         tree: Optional[TreeObject] = None,
-        retval: Optional[Dict[Any, Any]] = None,
+        retval: Optional[dict[Any, Any]] = None,
         fileobj: Optional[Any] = None,
-        stack: Optional[List[PdfObject]] = None,
-    ) -> Optional[Dict[str, Any]]:
+        stack: Optional[list[PdfObject]] = None,
+    ) -> Optional[dict[str, Any]]:
         """
         Extract field data if this PDF contains interactive form fields.
 
@@ -550,6 +544,7 @@ class PdfDocCommon:
             value is a :class:`Field<pypdf.generic.Field>` object. By
             default, the mapping name is used for keys.
             ``None`` if form data could not be located.
+
         """
         field_attributes = FA.attributes_dict()
         field_attributes.update(CheckboxRadioButtonAttributes.attributes_dict())
@@ -578,7 +573,7 @@ class PdfDocCommon:
     def _get_qualified_field_name(self, parent: DictionaryObject) -> str:
         if "/TM" in parent:
             return cast(str, parent["/TM"])
-        elif "/Parent" in parent:
+        if "/Parent" in parent:
             return (
                 self._get_qualified_field_name(
                     cast(DictionaryObject, parent["/Parent"])
@@ -586,16 +581,15 @@ class PdfDocCommon:
                 + "."
                 + cast(str, parent.get("/T", ""))
             )
-        else:
-            return cast(str, parent.get("/T", ""))
+        return cast(str, parent.get("/T", ""))
 
     def _build_field(
         self,
         field: Union[TreeObject, DictionaryObject],
-        retval: Dict[Any, Any],
+        retval: dict[Any, Any],
         fileobj: Any,
         field_attributes: Any,
-        stack: List[PdfObject],
+        stack: list[PdfObject],
     ) -> None:
         if all(attr not in field for attr in ("/T", "/TM")):
             return
@@ -605,7 +599,7 @@ class PdfDocCommon:
             fileobj.write("\n")
         retval[key] = Field(field)
         obj = retval[key].indirect_reference.get_object()  # to get the full object
-        if obj.get(FA.FT, "") == "/Ch":
+        if obj.get(FA.FT, "") == "/Ch" and obj.get(NameObject(FA.Opt)):
             retval[key][NameObject("/_States_")] = obj[NameObject(FA.Opt)]
         if obj.get(FA.FT, "") == "/Btn" and "/AP" in obj:
             #  Checkbox
@@ -615,7 +609,7 @@ class PdfDocCommon:
             if "/Off" not in retval[key]["/_States_"]:
                 retval[key][NameObject("/_States_")].append(NameObject("/Off"))
         elif obj.get(FA.FT, "") == "/Btn" and obj.get(FA.Ff, 0) & FA.FfBits.Radio != 0:
-            states: List[str] = []
+            states: list[str] = []
             retval[key][NameObject("/_States_")] = ArrayObject(states)
             for k in obj.get(FA.Kids, {}):
                 k = k.get_object()
@@ -636,7 +630,7 @@ class PdfDocCommon:
         tree: Union[TreeObject, DictionaryObject],
         retval: Any,
         fileobj: Any,
-        stack: List[PdfObject],
+        stack: list[PdfObject],
     ) -> None:
         if tree in stack:
             logger_warning(
@@ -644,9 +638,9 @@ class PdfDocCommon:
             )
             return
         stack.append(tree)
-        if PA.KIDS in tree:
+        if PagesAttributes.KIDS in tree:
             # recurse down the tree
-            for kid in tree[PA.KIDS]:  # type: ignore
+            for kid in tree[PagesAttributes.KIDS]:  # type: ignore
                 kid = kid.get_object()
                 self.get_fields(kid, retval, fileobj, stack)
 
@@ -665,7 +659,7 @@ class PdfDocCommon:
             attr_name = field_attributes[attr]
             try:
                 if attr == FA.FT:
-                    # Make the field type value more clear
+                    # Make the field type value clearer
                     types = {
                         "/Btn": "Button",
                         "/Tx": "Text",
@@ -687,7 +681,7 @@ class PdfDocCommon:
                 # Field attribute is N/A or unknown, so don't write anything
                 pass
 
-    def get_form_text_fields(self, full_qualified_name: bool = False) -> Dict[str, Any]:
+    def get_form_text_fields(self, full_qualified_name: bool = False) -> dict[str, Any]:
         """
         Retrieve form fields from the document with textual data.
 
@@ -700,17 +694,17 @@ class PdfDocCommon:
 
             If the document contains multiple form fields with the same name, the
             second and following will get the suffix .2, .3, ...
+
         """
 
-        def indexed_key(k: str, fields: Dict[Any, Any]) -> str:
+        def indexed_key(k: str, fields: dict[Any, Any]) -> str:
             if k not in fields:
                 return k
-            else:
-                return (
-                    k
-                    + "."
-                    + str(sum([1 for kk in fields if kk.startswith(k + ".")]) + 2)
-                )
+            return (
+                k
+                + "."
+                + str(sum(1 for kk in fields if kk.startswith(k + ".")) + 2)
+            )
 
         # Retrieve document form fields
         formfields = self.get_fields()
@@ -727,7 +721,7 @@ class PdfDocCommon:
 
     def get_pages_showing_field(
         self, field: Union[Field, PdfObject, IndirectObject]
-    ) -> List[PageObject]:
+    ) -> list[PageObject]:
         """
         Provides list of pages where the field is called.
 
@@ -745,25 +739,25 @@ class PdfDocCommon:
                 - Multi-page list:
                     Field with multiple kids widgets
                     (example: radio buttons, field repeated on multiple pages).
+
         """
 
         def _get_inherited(obj: DictionaryObject, key: str) -> Any:
             if key in obj:
                 return obj[key]
-            elif "/Parent" in obj:
+            if "/Parent" in obj:
                 return _get_inherited(
                     cast(DictionaryObject, obj["/Parent"].get_object()), key
                 )
-            else:
-                return None
+            return None
 
         try:
             # to cope with all types
             field = cast(DictionaryObject, field.indirect_reference.get_object())  # type: ignore
         except Exception as exc:
-            raise ValueError("field type is invalid") from exc
+            raise ValueError("Field type is invalid") from exc
         if is_null_or_none(_get_inherited(field, "/FT")):
-            raise ValueError("field is not valid")
+            raise ValueError("Field is not valid")
         ret = []
         if field.get("/Subtype", "") == "/Widget":
             if "/P" in field:
@@ -806,6 +800,7 @@ class PdfDocCommon:
 
         Raises:
             Exception: If a destination is invalid.
+
         """
         if "/OpenAction" not in self.root_object:
             return None
@@ -814,10 +809,9 @@ class PdfDocCommon:
             oa = oa.decode()
         if isinstance(oa, str):
             return create_string_object(oa)
-        elif isinstance(oa, ArrayObject):
+        if isinstance(oa, ArrayObject):
             try:
-                page, typ = oa[0:2]
-                array = oa[2:]
+                page, typ, *array = oa
                 fit = Fit(typ, tuple(array))
                 return Destination("OpenAction", page, fit)
             except Exception as exc:
@@ -827,7 +821,7 @@ class PdfDocCommon:
 
     @open_destination.setter
     def open_destination(self, dest: Union[None, str, Destination, PageObject]) -> None:
-        raise NotImplementedError("no setter for open_destination")
+        raise NotImplementedError("No setter for open_destination")
 
     @property
     def outline(self) -> OutlineType:
@@ -839,7 +833,10 @@ class PdfDocCommon:
         return self._get_outline()
 
     def _get_outline(
-        self, node: Optional[DictionaryObject] = None, outline: Optional[Any] = None
+        self,
+        node: Optional[DictionaryObject] = None,
+        outline: Optional[Any] = None,
+        visited: Optional[set[int]] = None,
     ) -> OutlineType:
         if outline is None:
             outline = []
@@ -855,21 +852,35 @@ class PdfDocCommon:
                 # §12.3.3 Document outline, entries in the outline dictionary
                 if not is_null_or_none(lines) and "/First" in lines:
                     node = cast(DictionaryObject, lines["/First"])
-            self._namedDests = self._get_named_destinations()
+            self._named_destinations = self._get_named_destinations()
 
         if node is None:
             return outline
 
         # see if there are any more outline items
+        if visited is None:
+            visited = set()
         while True:
+            node_id = id(node)
+            if node_id in visited:
+                logger_warning(f"Detected cycle in outline structure for {node}", __name__)
+                break
+            visited.add(node_id)
+
             outline_obj = self._build_outline_item(node)
             if outline_obj:
                 outline.append(outline_obj)
 
             # check for sub-outline
             if "/First" in node:
-                sub_outline: List[Any] = []
-                self._get_outline(cast(DictionaryObject, node["/First"]), sub_outline)
+                sub_outline: list[Any] = []
+                # Pass a copy to allow multiple outer entries to reference the same inner one.
+                inner_visited = visited.copy()
+                self._get_outline(
+                    node=cast(DictionaryObject, node["/First"]),
+                    outline=sub_outline,
+                    visited=inner_visited,
+                )
                 if sub_outline:
                     outline.append(sub_outline)
 
@@ -898,8 +909,7 @@ class PdfDocCommon:
         catalog = self.root_object
         if CO.THREADS in catalog:
             return cast("ArrayObject", catalog[CO.THREADS])
-        else:
-            return None
+        return None
 
     @abstractmethod
     def _get_page_number_by_indirect(
@@ -917,6 +927,7 @@ class PdfDocCommon:
 
         Returns:
             The page number or None if page is not found
+
         """
         return self._get_page_number_by_indirect(page.indirect_reference)
 
@@ -929,14 +940,15 @@ class PdfDocCommon:
 
         Returns:
             The page number or None if page is not found
+
         """
         return self._get_page_number_by_indirect(destination.page)
 
     def _build_destination(
         self,
-        title: str,
+        title: Union[str, bytes],
         array: Optional[
-            List[
+            list[
                 Union[NumberObject, IndirectObject, None, NullObject, DictionaryObject]
             ]
         ],
@@ -950,25 +962,23 @@ class PdfDocCommon:
         ):
             page = NullObject()
             return Destination(title, page, Fit.fit())
-        else:
-            page, typ = array[0:2]  # type: ignore
-            array = array[2:]
-            try:
-                return Destination(title, page, Fit(fit_type=typ, fit_args=array))  # type: ignore
-            except PdfReadError:
-                logger_warning(f"Unknown destination: {title} {array}", __name__)
-                if self.strict:
-                    raise
-                # create a link to first Page
-                tmp = self.pages[0].indirect_reference
-                indirect_reference = NullObject() if tmp is None else tmp
-                return Destination(title, indirect_reference, Fit.fit())  # type: ignore
+        page, typ, *array = array  # type: ignore
+        try:
+            return Destination(title, page, Fit(fit_type=typ, fit_args=array))  # type: ignore
+        except PdfReadError:
+            logger_warning(f"Unknown destination: {title!r} {array}", __name__)
+            if self.strict:
+                raise
+            # create a link to first Page
+            tmp = self.pages[0].indirect_reference
+            indirect_reference = NullObject() if tmp is None else tmp
+            return Destination(title, indirect_reference, Fit.fit())
 
     def _build_outline_item(self, node: DictionaryObject) -> Optional[Destination]:
         dest, title, outline_item = None, None, None
 
         # title required for valid outline
-        # § 12.3.3, entries in an outline item dictionary
+        # §12.3.3, entries in an outline item dictionary
         try:
             title = cast("str", node["/Title"])
         except KeyError:
@@ -977,13 +987,16 @@ class PdfDocCommon:
             title = ""
 
         if "/A" in node:
-            # Action, PDFv1.7 Section 12.6 (only type GoTo supported)
+            # Action, PDF 1.7 and PDF 2.0 §12.6 (only type GoTo supported)
             action = cast(DictionaryObject, node["/A"])
             action_type = cast(NameObject, action[GoToActionArguments.S])
             if action_type == "/GoTo":
-                dest = action[GoToActionArguments.D]
+                if GoToActionArguments.D in action:
+                    dest = action[GoToActionArguments.D]
+                elif self.strict:
+                    raise PdfReadError(f"Outline Action Missing /D attribute: {node!r}")
         elif "/Dest" in node:
-            # Destination, PDFv1.7 Section 12.3.2
+            # Destination, PDF 1.7 and PDF 2.0 §12.3.2
             dest = node["/Dest"]
             # if array was referenced in another object, will be a dict w/ key "/D"
             if isinstance(dest, DictionaryObject) and "/D" in dest:
@@ -993,10 +1006,10 @@ class PdfDocCommon:
             outline_item = self._build_destination(title, dest)
         elif isinstance(dest, str):
             # named destination, addresses NameObject Issue #193
-            # TODO : keep named destination instead of replacing it ?
+            # TODO: Keep named destination instead of replacing it?
             try:
                 outline_item = self._build_destination(
-                    title, self._namedDests[dest].dest_array
+                    title, self._named_destinations[dest].dest_array
                 )
             except KeyError:
                 # named destination not found in Name Dict
@@ -1008,11 +1021,10 @@ class PdfDocCommon:
         else:
             if self.strict:
                 raise PdfReadError(f"Unexpected destination {dest!r}")
-            else:
-                logger_warning(
-                    f"Removed unexpected destination {dest!r} from destination",
-                    __name__,
-                )
+            logger_warning(
+                f"Removed unexpected destination {dest!r} from destination",
+                __name__,
+            )
             outline_item = self._build_destination(title, None)
 
         # if outline item created, add color, format, and child count if present
@@ -1028,7 +1040,7 @@ class PdfDocCommon:
                 # absolute value = num. visible children
                 # with positive = open/unfolded, negative = closed/folded
                 outline_item[NameObject("/Count")] = node["/Count"]
-            #  if count is 0 we will consider it as open ( in order to have always an is_open to simplify
+            #  if count is 0 we will consider it as open (to have available is_open)
             outline_item[NameObject("/%is_open%")] = BooleanObject(
                 node.get("/Count", 0) >= 0
             )
@@ -1040,7 +1052,7 @@ class PdfDocCommon:
         return outline_item
 
     @property
-    def pages(self) -> List[PageObject]:
+    def pages(self) -> list[PageObject]:
         """
         Property that emulates a list of :class:`PageObject<pypdf._page.PageObject>`.
         This property allows to get a page or a range of pages.
@@ -1057,7 +1069,7 @@ class PdfDocCommon:
         return _VirtualList(self.get_num_pages, self.get_page)  # type: ignore
 
     @property
-    def page_labels(self) -> List[str]:
+    def page_labels(self) -> list[str]:
         """
         A list of labels for the pages in this document.
 
@@ -1124,17 +1136,27 @@ class PdfDocCommon:
         self,
         list_only: bool = False,
         pages: Union[None, DictionaryObject, PageObject] = None,
-        inherit: Optional[Dict[str, Any]] = None,
+        inherit: Optional[dict[str, Any]] = None,
         indirect_reference: Optional[IndirectObject] = None,
     ) -> None:
         """
-        Prepare the document pages to ease searching
+        Process the document pages to ease searching.
+
+        Attributes of a page may inherit from ancestor nodes
+        in the page tree. Flattening means moving
+        any inheritance data into descendant nodes,
+        effectively removing the inheritance dependency.
+
+        Note: It is distinct from another use of "flattening" applied to PDFs.
+        Flattening a PDF also means combining all the contents into one single layer
+        and making the file less editable.
 
         Args:
             list_only: Will only list the pages within _flatten_pages.
             pages:
             inherit:
             indirect_reference: Used recursively to flatten the /Pages object.
+
         """
         inheritable_page_attributes = (
             NameObject(PG.RESOURCES),
@@ -1148,14 +1170,15 @@ class PdfDocCommon:
             # Fix issue 327: set flattened_pages attribute only for
             # decrypted file
             catalog = self.root_object
-            pages = catalog["/Pages"].get_object()  # type: ignore
-            assert isinstance(pages, DictionaryObject)
+            pages = catalog.get("/Pages").get_object()  # type: ignore
+            if not isinstance(pages, DictionaryObject):
+                raise PdfReadError("Invalid object in /Pages")
             self.flattened_pages = []
 
-        if PA.TYPE in pages:
-            t = cast(str, pages[PA.TYPE])
-        # if pdf has no type, considered as a page if /Kids is missing
-        elif PA.KIDS not in pages:
+        if PagesAttributes.TYPE in pages:
+            t = cast(str, pages[PagesAttributes.TYPE])
+        # if the page tree node has no /Type, consider as a page if /Kids is also missing
+        elif PagesAttributes.KIDS not in pages:
             t = "/Page"
         else:
             t = "/Pages"
@@ -1164,7 +1187,11 @@ class PdfDocCommon:
             for attr in inheritable_page_attributes:
                 if attr in pages:
                     inherit[attr] = pages[attr]
-            for page in cast(ArrayObject, pages[PA.KIDS]):
+            pages_reference = getattr(pages, "indirect_reference", object())
+            for page in cast(ArrayObject, pages[PagesAttributes.KIDS]):
+                if getattr(page, "indirect_reference", object()) == pages_reference:
+                    raise PdfReadError("Detected cyclic page references.")
+
                 addt = {}
                 if isinstance(page, IndirectObject):
                     addt["indirect_reference"] = page
@@ -1178,9 +1205,9 @@ class PdfDocCommon:
                             "Maximum recursion depth reached during page flattening."
                         )
         elif t == "/Page":
-            for attr_in, value in list(inherit.items()):
-                # if the page has it's own value, it does not inherit the
-                # parent's value:
+            for attr_in, value in inherit.items():
+                # if the page has its own value, it does not inherit the
+                # parent's value
                 if attr_in not in pages:
                     pages[attr_in] = value
             page_obj = PageObject(self, indirect_reference)
@@ -1207,6 +1234,7 @@ class PdfDocCommon:
 
             clean: replace PageObject with NullObject to prevent annotations
                 or destinations to reference a detached page.
+
         """
         if self.flattened_pages is None:
             self._flatten(self._readonly)
@@ -1245,14 +1273,15 @@ class PdfDocCommon:
 
         Returns:
             A PdfObject
+
         """
         return IndirectObject(num, gen, self).get_object()
 
     def decode_permissions(
         self, permissions_code: int
-    ) -> Dict[str, bool]:  # pragma: no cover
+    ) -> dict[str, bool]:  # pragma: no cover
         """Take the permissions as an integer, return the allowed access."""
-        deprecate_with_replacement(
+        deprecation_with_replacement(
             old_name="decode_permissions",
             new_name="user_access_permissions",
             removed_in="5.0.0",
@@ -1294,9 +1323,9 @@ class PdfDocCommon:
         ...  # pragma: no cover
 
     @property
-    def xfa(self) -> Optional[Dict[str, Any]]:
+    def xfa(self) -> Optional[dict[str, Any]]:
         tree: Optional[TreeObject] = None
-        retval: Dict[str, Any] = {}
+        retval: dict[str, Any] = {}
         catalog = self.root_object
 
         if "/AcroForm" not in catalog or not catalog["/AcroForm"]:
@@ -1318,7 +1347,8 @@ class PdfDocCommon:
         return retval
 
     @property
-    def attachments(self) -> Mapping[str, List[bytes]]:
+    def attachments(self) -> Mapping[str, list[bytes]]:
+        """Mapping of attachment filenames to their content."""
         return LazyDict(
             {
                 name: (self._get_attachment_list, name)
@@ -1326,29 +1356,27 @@ class PdfDocCommon:
             }
         )
 
-    def _list_attachments(self) -> List[str]:
+    @property
+    def attachment_list(self) -> Generator[EmbeddedFile, None, None]:
+        """Iterable of attachment objects."""
+        yield from EmbeddedFile._load(self.root_object)
+
+    def _list_attachments(self) -> list[str]:
         """
         Retrieves the list of filenames of file attachments.
 
         Returns:
             list of filenames
-        """
-        catalog = self.root_object
-        # From the catalog get the embedded file names
-        try:
-            filenames = cast(
-                ArrayObject,
-                cast(
-                    DictionaryObject,
-                    cast(DictionaryObject, catalog["/Names"])["/EmbeddedFiles"],
-                )["/Names"],
-            )
-        except KeyError:
-            return []
-        attachments_names = [f for f in filenames if isinstance(f, str)]
-        return attachments_names
 
-    def _get_attachment_list(self, name: str) -> List[bytes]:
+        """
+        names = []
+        for entry in self.attachment_list:
+            names.append(entry.name)
+            if (name := entry.alternative_name) != entry.name and name:
+                names.append(name)
+        return names
+
+    def _get_attachment_list(self, name: str) -> list[bytes]:
         out = self._get_attachments(name)[name]
         if isinstance(out, list):
             return out
@@ -1356,7 +1384,7 @@ class PdfDocCommon:
 
     def _get_attachments(
         self, filename: Optional[str] = None
-    ) -> Dict[str, Union[bytes, List[bytes]]]:
+    ) -> dict[str, Union[bytes, list[bytes]]]:
         """
         Retrieves all or selected file attachments of the PDF as a dictionary of file names
         and the file data as a bytestring.
@@ -1370,41 +1398,54 @@ class PdfDocCommon:
         Returns:
             dictionary of filename -> Union[bytestring or List[ByteString]]
             If the filename exists multiple times a list of the different versions will be provided.
+
         """
-        catalog = self.root_object
-        # From the catalog get the embedded file names
-        try:
-            filenames = cast(
-                ArrayObject,
-                cast(
-                    DictionaryObject,
-                    cast(DictionaryObject, catalog["/Names"])["/EmbeddedFiles"],
-                )["/Names"],
-            )
-        except KeyError:
-            return {}
-        attachments: Dict[str, Union[bytes, List[bytes]]] = {}
-        # Loop through attachments
-        for i in range(len(filenames)):
-            f = filenames[i]
-            if isinstance(f, str):
-                if filename is not None and f != filename:
+        attachments: dict[str, Union[bytes, list[bytes]]] = {}
+        for entry in self.attachment_list:
+            names = set()
+            alternative_name = entry.alternative_name
+            if filename is not None:
+                if filename in {entry.name, alternative_name}:
+                    name = entry.name if filename == entry.name else alternative_name
+                    names.add(name)
+                else:
+                    continue
+            else:
+                names = {entry.name, alternative_name}
+
+            for name in names:
+                if name is None:
                     continue
-                name = f
-                f_dict = filenames[i + 1].get_object()
-                f_data = f_dict["/EF"]["/F"].get_data()
                 if name in attachments:
                     if not isinstance(attachments[name], list):
                         attachments[name] = [attachments[name]]  # type:ignore
-                    attachments[name].append(f_data)  # type:ignore
+                    attachments[name].append(entry.content)  # type:ignore
                 else:
-                    attachments[name] = f_data
+                    attachments[name] = entry.content
         return attachments
 
+    @abstractmethod
+    def _repr_mimebundle_(
+        self,
+        include: Union[None, Iterable[str]] = None,
+        exclude: Union[None, Iterable[str]] = None,
+    ) -> dict[str, Any]:
+        """
+        Integration into Jupyter Notebooks.
+
+        This method returns a dictionary that maps a mime-type to its
+        representation.
+
+        .. seealso::
+
+            https://ipython.readthedocs.io/en/stable/config/integrating.html
+        """
+        ...  # pragma: no cover
+
 
 class LazyDict(Mapping[Any, Any]):
-    def __init__(self, *args: Any, **kw: Any) -> None:
-        self._raw_dict = dict(*args, **kw)
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
+        self._raw_dict = dict(*args, **kwargs)
 
     def __getitem__(self, key: str) -> Any:
         func, arg = self._raw_dict.__getitem__(key)
